"""Tests for tescmd.api.command — CommandAPI."""

from __future__ import annotations

import json
from typing import TYPE_CHECKING

import pytest

from tescmd.api.command import CommandAPI
from tescmd.models.command import CommandResponse

if TYPE_CHECKING:
    from pytest_httpx import HTTPXMock

    from tescmd.api.client import TeslaFleetClient

FLEET_BASE = "https://fleet-api.prd.na.vn.cloud.tesla.com"
VIN = "5YJ3E1EA1NF000001"

_OK_RESPONSE = {"response": {"result": True, "reason": ""}}
_FAIL_RESPONSE = {"response": {"result": False, "reason": "vehicle is asleep"}}


def _body(mock: HTTPXMock, idx: int = 0) -> dict[str, object]:
    """Parse the JSON body of the idx-th request."""
    return json.loads(mock.get_requests()[idx].content)  # type: ignore[no-any-return]


class TestChargeCommands:
    @pytest.mark.asyncio
    async def test_charge_start(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(
            url=f"{FLEET_BASE}/api/1/vehicles/{VIN}/command/charge_start",
            json=_OK_RESPONSE,
        )
        api = CommandAPI(mock_client)
        result = await api.charge_start(VIN)

        assert isinstance(result, CommandResponse)
        assert result.response.result is True

    @pytest.mark.asyncio
    async def test_set_charge_limit_sends_body(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.set_charge_limit(VIN, percent=80)

        request = httpx_mock.get_requests()[0]
        assert request.url.path == f"/api/1/vehicles/{VIN}/command/set_charge_limit"
        assert _body(httpx_mock)["percent"] == 80

    @pytest.mark.asyncio
    async def test_set_charging_amps_sends_body(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.set_charging_amps(VIN, charging_amps=32)

        assert _body(httpx_mock)["charging_amps"] == 32

    @pytest.mark.asyncio
    async def test_set_scheduled_charging_sends_body(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.set_scheduled_charging(VIN, enable=True, time=480)

        body = _body(httpx_mock)
        assert body["enable"] is True
        assert body["time"] == 480

    @pytest.mark.asyncio
    async def test_command_failure_response(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_FAIL_RESPONSE)
        api = CommandAPI(mock_client)
        result = await api.charge_start(VIN)

        assert result.response.result is False
        assert result.response.reason == "vehicle is asleep"


class TestClimateCommands:
    @pytest.mark.asyncio
    async def test_set_temps_sends_body(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.set_temps(VIN, driver_temp=22.0, passenger_temp=20.0)

        request = httpx_mock.get_requests()[0]
        assert request.url.path == f"/api/1/vehicles/{VIN}/command/set_temps"
        body = _body(httpx_mock)
        assert body["driver_temp"] == 22.0
        assert body["passenger_temp"] == 20.0

    @pytest.mark.asyncio
    async def test_remote_seat_heater_request(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.remote_seat_heater_request(VIN, seat_position=0, level=3)

        body = _body(httpx_mock)
        assert body["seat_position"] == 0
        assert body["level"] == 3

    @pytest.mark.asyncio
    async def test_set_climate_keeper_mode(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.set_climate_keeper_mode(VIN, climate_keeper_mode=2)

        assert _body(httpx_mock)["climate_keeper_mode"] == 2


class TestCabinOverheatProtection:
    @pytest.mark.asyncio
    async def test_fan_only_defaults_false(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.set_cabin_overheat_protection(VIN, on=True)

        body = _body(httpx_mock)
        assert body["on"] is True
        assert body["fan_only"] is False

    @pytest.mark.asyncio
    async def test_fan_only_explicit_true(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.set_cabin_overheat_protection(VIN, on=True, fan_only=True)

        body = _body(httpx_mock)
        assert body["on"] is True
        assert body["fan_only"] is True


class TestSecurityCommands:
    @pytest.mark.asyncio
    async def test_door_lock_no_body(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(
            url=f"{FLEET_BASE}/api/1/vehicles/{VIN}/command/door_lock",
            json=_OK_RESPONSE,
        )
        api = CommandAPI(mock_client)
        result = await api.door_lock(VIN)

        assert result.response.result is True
        request = httpx_mock.get_requests()[0]
        # No body for lock
        assert request.content == b""

    @pytest.mark.asyncio
    async def test_set_sentry_mode_sends_body(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.set_sentry_mode(VIN, on=True)

        assert _body(httpx_mock)["on"] is True

    @pytest.mark.asyncio
    async def test_set_valet_mode_with_password(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.set_valet_mode(VIN, on=True, password="1234")

        body = _body(httpx_mock)
        assert body["on"] is True
        assert body["password"] == "1234"

    @pytest.mark.asyncio
    async def test_speed_limit_activate(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.speed_limit_activate(VIN, pin="1234")

        assert _body(httpx_mock)["pin"] == "1234"


class TestTrunkCommands:
    @pytest.mark.asyncio
    async def test_actuate_trunk_sends_which(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.actuate_trunk(VIN, which_trunk="rear")

        request = httpx_mock.get_requests()[0]
        assert request.url.path == f"/api/1/vehicles/{VIN}/command/actuate_trunk"
        assert _body(httpx_mock)["which_trunk"] == "rear"

    @pytest.mark.asyncio
    async def test_window_control_sends_params(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.window_control(VIN, command="vent", lat=0.0, lon=0.0)

        body = _body(httpx_mock)
        assert body["command"] == "vent"
        assert body["lat"] == 0.0
        assert body["lon"] == 0.0


# ------------------------------------------------------------------
# Phase 4: New command tests
# ------------------------------------------------------------------


class TestHomelinkCommands:
    @pytest.mark.asyncio
    async def test_trigger_homelink(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        result = await api.trigger_homelink(VIN, lat=37.77, lon=-122.42)

        assert result.response.result is True
        body = _body(httpx_mock)
        assert body["lat"] == 37.77
        assert body["lon"] == -122.42


class TestScheduledDeparture:
    @pytest.mark.asyncio
    async def test_set_scheduled_departure(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.set_scheduled_departure(
            VIN,
            enable=True,
            departure_time=480,
            preconditioning_enabled=True,
            off_peak_charging_enabled=True,
            end_off_peak_time=360,
        )

        body = _body(httpx_mock)
        assert body["enable"] is True
        assert body["departure_time"] == 480
        assert body["preconditioning_enabled"] is True
        assert body["off_peak_charging_enabled"] is True
        assert body["end_off_peak_time"] == 360


class TestChargeSchedule:
    @pytest.mark.asyncio
    async def test_add_charge_schedule_with_params(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.add_charge_schedule(
            VIN, id=1, enabled=True, start_time=360, end_time=480, days_of_week="0111110"
        )

        body = _body(httpx_mock)
        assert body["id"] == 1
        assert body["enabled"] is True
        assert body["start_time"] == 360
        assert body["days_of_week"] == "0111110"

    @pytest.mark.asyncio
    async def test_add_charge_schedule_omits_none_params(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.add_charge_schedule(VIN, id=1, enabled=True)

        body = _body(httpx_mock)
        assert body == {"id": 1, "enabled": True}

    @pytest.mark.asyncio
    async def test_remove_charge_schedule(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.remove_charge_schedule(VIN, id=1)

        assert _body(httpx_mock)["id"] == 1


class TestPreconditionSchedule:
    @pytest.mark.asyncio
    async def test_add_precondition_schedule_with_params(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.add_precondition_schedule(VIN, id=1, enabled=True, days_of_week="1111111")

        body = _body(httpx_mock)
        assert body["id"] == 1
        assert body["enabled"] is True
        assert body["days_of_week"] == "1111111"

    @pytest.mark.asyncio
    async def test_add_precondition_schedule_omits_none_params(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.add_precondition_schedule(VIN, id=2, enabled=False)

        body = _body(httpx_mock)
        assert body == {"id": 2, "enabled": False}

    @pytest.mark.asyncio
    async def test_remove_precondition_schedule(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.remove_precondition_schedule(VIN, id=1)

        assert _body(httpx_mock)["id"] == 1


class TestClimateNewCommands:
    @pytest.mark.asyncio
    async def test_set_cop_temp(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.set_cop_temp(VIN, cop_temp=2)

        assert _body(httpx_mock)["cop_temp"] == 2

    @pytest.mark.asyncio
    async def test_remote_auto_seat_climate(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.remote_auto_seat_climate_request(VIN, auto_seat_position=0, auto_climate_on=True)

        body = _body(httpx_mock)
        assert body["auto_seat_position"] == 0
        assert body["auto_climate_on"] is True

    @pytest.mark.asyncio
    async def test_remote_auto_steering_wheel_heat(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.remote_auto_steering_wheel_heat_climate_request(VIN, on=True)

        assert _body(httpx_mock)["on"] is True

    @pytest.mark.asyncio
    async def test_steering_wheel_heat_level(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.remote_steering_wheel_heat_level_request(VIN, level=3)

        assert _body(httpx_mock)["level"] == 3


class TestSecurityNewCommands:
    @pytest.mark.asyncio
    async def test_reset_pin_to_drive_pin(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        result = await api.reset_pin_to_drive_pin(VIN)

        assert result.response.result is True
        request = httpx_mock.get_requests()[0]
        assert request.content == b""

    @pytest.mark.asyncio
    async def test_clear_pin_to_drive_admin(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        result = await api.clear_pin_to_drive_admin(VIN)

        assert result.response.result is True

    @pytest.mark.asyncio
    async def test_speed_limit_clear_pin(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.speed_limit_clear_pin(VIN, pin="1234")

        assert _body(httpx_mock)["pin"] == "1234"

    @pytest.mark.asyncio
    async def test_speed_limit_clear_pin_admin(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        result = await api.speed_limit_clear_pin_admin(VIN)

        assert result.response.result is True


class TestNavigationNewCommands:
    @pytest.mark.asyncio
    async def test_navigation_waypoints(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        waypoints_str = "refId:ChIJIQBpAG2ahYAR_6128GcTUEo,refId:ChIJw____96GhYARCVVwg5cT7c0"
        await api.navigation_waypoints_request(VIN, waypoints=waypoints_str)

        body = _body(httpx_mock)
        assert body["waypoints"] == waypoints_str
        assert "navigation_waypoints_request" in str(httpx_mock.get_requests()[0].url)


class TestNavigationScRequest:
    @pytest.mark.asyncio
    async def test_navigation_sc_request_with_params(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.navigation_sc_request(VIN, id=42, order=1)

        body = _body(httpx_mock)
        assert body["id"] == 42
        assert body["order"] == 1
        assert "navigation_sc_request" in str(httpx_mock.get_requests()[0].url)

    @pytest.mark.asyncio
    async def test_navigation_sc_request_defaults(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.navigation_sc_request(VIN)

        body = _body(httpx_mock)
        assert body["id"] == 0
        assert body["order"] == 0


class TestBoomboxCommand:
    @pytest.mark.asyncio
    async def test_remote_boombox_default_sound(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        result = await api.remote_boombox(VIN)

        assert result.response.result is True
        body = _body(httpx_mock)
        assert body["sound"] == 2000

    @pytest.mark.asyncio
    async def test_remote_boombox_fart_sound(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.remote_boombox(VIN, sound=0)

        body = _body(httpx_mock)
        assert body["sound"] == 0


class TestSpeedLimitFloat:
    @pytest.mark.asyncio
    async def test_speed_limit_set_limit_float(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.speed_limit_set_limit(VIN, limit_mph=65.5)

        body = _body(httpx_mock)
        assert body["limit_mph"] == 65.5

    @pytest.mark.asyncio
    async def test_speed_limit_set_limit_int_still_works(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.speed_limit_set_limit(VIN, limit_mph=65)

        body = _body(httpx_mock)
        assert body["limit_mph"] == 65


class TestTonneauCommands:
    @pytest.mark.asyncio
    async def test_open_tonneau(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        result = await api.open_tonneau(VIN)

        assert result.response.result is True
        request = httpx_mock.get_requests()[0]
        assert "open_tonneau" in str(request.url)
        assert request.content == b""

    @pytest.mark.asyncio
    async def test_close_tonneau(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        result = await api.close_tonneau(VIN)

        assert result.response.result is True
        assert "close_tonneau" in str(httpx_mock.get_requests()[0].url)

    @pytest.mark.asyncio
    async def test_stop_tonneau(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        result = await api.stop_tonneau(VIN)

        assert result.response.result is True
        assert "stop_tonneau" in str(httpx_mock.get_requests()[0].url)


class TestPowerManagementCommands:
    @pytest.mark.asyncio
    async def test_set_low_power_mode_on(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        result = await api.set_low_power_mode(VIN, enable=True)

        assert result.response.result is True
        assert _body(httpx_mock)["enable"] is True

    @pytest.mark.asyncio
    async def test_set_low_power_mode_off(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.set_low_power_mode(VIN, enable=False)

        assert _body(httpx_mock)["enable"] is False

    @pytest.mark.asyncio
    async def test_keep_accessory_power_mode_on(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        result = await api.keep_accessory_power_mode(VIN, enable=True)

        assert result.response.result is True
        assert _body(httpx_mock)["enable"] is True

    @pytest.mark.asyncio
    async def test_keep_accessory_power_mode_off(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.keep_accessory_power_mode(VIN, enable=False)

        assert _body(httpx_mock)["enable"] is False


class TestManagedChargingCommand:
    @pytest.mark.asyncio
    async def test_set_managed_charge_current_request(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        result = await api.set_managed_charge_current_request(VIN, charging_amps=16)

        assert result.response.result is True
        body = _body(httpx_mock)
        assert body["charging_amps"] == 16
        assert "set_managed_charge_current_request" in str(httpx_mock.get_requests()[0].url)


class TestNavigationRequest:
    @pytest.mark.asyncio
    async def test_navigation_request_sends_correct_body(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        result = await api.navigation_request(
            VIN, value={"android.intent.extra.TEXT": "123 Main St"}
        )

        assert result.response.result is True
        body = _body(httpx_mock)
        assert body["type"] == "share_ext_content_raw"
        assert body["locale"] == "en-US"
        assert "timestamp_ms" in body
        assert body["value"] == {"android.intent.extra.TEXT": "123 Main St"}
        request = httpx_mock.get_requests()[0]
        assert request.url.path == f"/api/1/vehicles/{VIN}/command/navigation_request"


class TestManagedChargingExtended:
    @pytest.mark.asyncio
    async def test_set_managed_charger_location(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        result = await api.set_managed_charger_location(
            VIN, location={"lat": 37.77, "lon": -122.42}
        )

        assert result.response.result is True
        body = _body(httpx_mock)
        assert body["lat"] == 37.77
        assert body["lon"] == -122.42
        assert "set_managed_charger_location" in str(httpx_mock.get_requests()[0].url)

    @pytest.mark.asyncio
    async def test_set_managed_scheduled_charging_time(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        result = await api.set_managed_scheduled_charging_time(VIN, time=480)

        assert result.response.result is True
        assert _body(httpx_mock) == {"time": 480}


class TestAdjustVolumeFloat:
    @pytest.mark.asyncio
    async def test_adjust_volume_sends_float(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        await api.adjust_volume(VIN, volume=5.5)

        body = _body(httpx_mock)
        assert body["volume"] == 5.5
        assert isinstance(body["volume"], float)


class TestPreconditioningManualOverride:
    @pytest.mark.asyncio
    async def test_set_preconditioning_max_with_manual_override(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        result = await api.set_preconditioning_max(VIN, on=True, manual_override=True)

        assert result.response.result is True
        body = _body(httpx_mock)
        assert body["on"] is True
        assert body["manual_override"] is True


class TestClimateKeeperManualOverride:
    @pytest.mark.asyncio
    async def test_set_climate_keeper_mode_with_manual_override(
        self, httpx_mock: HTTPXMock, mock_client: TeslaFleetClient
    ) -> None:
        httpx_mock.add_response(json=_OK_RESPONSE)
        api = CommandAPI(mock_client)
        result = await api.set_climate_keeper_mode(
            VIN, climate_keeper_mode=2, manual_override=True
        )

        assert result.response.result is True
        body = _body(httpx_mock)
        assert body["climate_keeper_mode"] == 2
        assert body["manual_override"] is True



"""Tests for DualGateFilter and haversine distance."""

from __future__ import annotations

from tescmd.openclaw.config import FieldFilter
from tescmd.openclaw.filters import DualGateFilter, haversine


class TestHaversine:
    def test_same_point_is_zero(self) -> None:
        assert haversine(40.0, -74.0, 40.0, -74.0) == 0.0

    def test_known_distance(self) -> None:
        # New York (40.7128, -74.0060) to Los Angeles (34.0522, -118.2437)
        # ~3944 km
        dist = haversine(40.7128, -74.0060, 34.0522, -118.2437)
        assert 3_900_000 < dist < 4_000_000  # meters

    def test_short_distance(self) -> None:
        # Two points ~111m apart (0.001 degrees latitude at equator)
        dist = haversine(0.0, 0.0, 0.001, 0.0)
        assert 100 < dist < 120  # ~111m

    def test_antipodal_points(self) -> None:
        # Opposite sides of Earth: ~20,000 km
        dist = haversine(0.0, 0.0, 0.0, 180.0)
        assert 20_000_000 < dist < 20_100_000


class TestDualGateFilter:
    def _make_filter(self, **fields: tuple[float, float]) -> DualGateFilter:
        """Helper: create filter from field_name → (granularity, throttle_seconds)."""
        filters = {}
        for name, (gran, throttle) in fields.items():
            filters[name] = FieldFilter(granularity=gran, throttle_seconds=throttle)
        return DualGateFilter(filters)

    def test_first_value_always_emits(self) -> None:
        filt = self._make_filter(Soc=(5.0, 10.0))
        assert filt.should_emit("Soc", 72, 0.0) is True

    def test_unknown_field_uses_default_filter(self) -> None:
        """Unconfigured fields pass through with the module-level default fallback."""
        filt = self._make_filter(Soc=(5.0, 10.0))
        # First value for an unconfigured field — always emits
        assert filt.should_emit("Unknown", 42, 0.0) is True

    def test_default_filter_throttles(self) -> None:
        """Default fallback enforces its 5-second throttle on unconfigured fields."""
        filt = self._make_filter()
        assert filt.should_emit("NewField", 1, 0.0) is True
        filt.record_emit("NewField", 1, 0.0)
        # Changed value but within 5s throttle window
        assert filt.should_emit("NewField", 2, 3.0) is False
        # After 5s throttle
        assert filt.should_emit("NewField", 2, 6.0) is True

    def test_disabled_field_still_rejected(self) -> None:
        """Explicitly disabled fields are blocked even if unconfigured ones pass."""
        filters = {"Blocked": FieldFilter(enabled=False, granularity=0.0, throttle_seconds=0.0)}
        filt = DualGateFilter(filters)
        assert filt.should_emit("Blocked", 42, 0.0) is False

    def test_disabled_field_rejected(self) -> None:
        filters = {"Soc": FieldFilter(enabled=False, granularity=0.0, throttle_seconds=0.0)}
        filt = DualGateFilter(filters)
        assert filt.should_emit("Soc", 72, 0.0) is False

    def test_delta_below_threshold_rejected(self) -> None:
        filt = self._make_filter(Soc=(5.0, 0.0))
        assert filt.should_emit("Soc", 72, 0.0) is True
        filt.record_emit("Soc", 72, 0.0)
        # Change of 2, below granularity of 5
        assert filt.should_emit("Soc", 74, 1.0) is False

    def test_delta_at_threshold_emits(self) -> None:
        filt = self._make_filter(Soc=(5.0, 0.0))
        assert filt.should_emit("Soc", 72, 0.0) is True
        filt.record_emit("Soc", 72, 0.0)
        # Change of 5, equal to granularity
        assert filt.should_emit("Soc", 77, 1.0) is True

    def test_delta_above_threshold_emits(self) -> None:
        filt = self._make_filter(Soc=(5.0, 0.0))
        assert filt.should_emit("Soc", 72, 0.0) is True
        filt.record_emit("Soc", 72, 0.0)
        assert filt.should_emit("Soc", 80, 1.0) is True

    def test_throttle_blocks_rapid_emissions(self) -> None:
        filt = self._make_filter(Soc=(0.0, 10.0))
        assert filt.should_emit("Soc", 72, 0.0) is True
        filt.record_emit("Soc", 72, 0.0)
        # Value changed, but throttle (10s) not elapsed
        assert filt.should_emit("Soc", 80, 5.0) is False

    def test_throttle_allows_after_interval(self) -> None:
        filt = self._make_filter(Soc=(0.0, 10.0))
        assert filt.should_emit("Soc", 72, 0.0) is True
        filt.record_emit("Soc", 72, 0.0)
        # Throttle elapsed, value changed
        assert filt.should_emit("Soc", 80, 11.0) is True

    def test_zero_granularity_emits_on_any_change(self) -> None:
        filt = self._make_filter(ChargeState=(0.0, 0.0))
        assert filt.should_emit("ChargeState", "Charging", 0.0) is True
        filt.record_emit("ChargeState", "Charging", 0.0)
        # Same value — should NOT emit
        assert filt.should_emit("ChargeState", "Charging", 1.0) is False
        # Different value — should emit
        assert filt.should_emit("ChargeState", "Complete", 1.0) is True

    def test_location_uses_haversine(self) -> None:
        filt = self._make_filter(Location=(50.0, 0.0))  # 50m granularity
        loc1 = {"latitude": 40.7128, "longitude": -74.0060}
        loc2 = {"latitude": 40.7128, "longitude": -74.0060}  # same
        loc3 = {"latitude": 40.7138, "longitude": -74.0060}  # ~111m away

        assert filt.should_emit("Location", loc1, 0.0) is True
        filt.record_emit("Location", loc1, 0.0)

        # Same location — below 50m threshold
        assert filt.should_emit("Location", loc2, 1.0) is False

        # ~111m away — above 50m threshold
        assert filt.should_emit("Location", loc3, 1.0) is True

    def test_both_gates_must_pass(self) -> None:
        """Delta passes but throttle blocks."""
        filt = self._make_filter(Soc=(5.0, 10.0))
        assert filt.should_emit("Soc", 72, 0.0) is True
        filt.record_emit("Soc", 72, 0.0)
        # Big delta (10) but throttle not elapsed (5s < 10s)
        assert filt.should_emit("Soc", 82, 5.0) is False
        # Throttle elapsed (11s > 10s)
        assert filt.should_emit("Soc", 82, 11.0) is True

    def test_reset_clears_state(self) -> None:
        filt = self._make_filter(Soc=(5.0, 10.0))
        filt.record_emit("Soc", 72, 0.0)
        filt.reset()
        # After reset, first value always emits
        assert filt.should_emit("Soc", 72, 0.0) is True

    def test_non_numeric_value_treated_as_changed(self) -> None:
        filt = self._make_filter(Gear=(1.0, 0.0))
        assert filt.should_emit("Gear", "D", 0.0) is True
        filt.record_emit("Gear", "D", 0.0)
        # Non-numeric delta → infinity → always passes
        assert filt.should_emit("Gear", "R", 1.0) is True


class TestStalenessGate:
    """Tests for max_seconds (staleness override)."""

    def test_max_seconds_forces_emit_when_stale(self) -> None:
        """Value barely changes but max_seconds elapsed → force emit."""
        filters = {"Soc": FieldFilter(granularity=5.0, throttle_seconds=1.0, max_seconds=60.0)}
        filt = DualGateFilter(filters)
        assert filt.should_emit("Soc", 72, 0.0) is True
        filt.record_emit("Soc", 72, 0.0)
        # Delta too small (1), throttle passed, but max_seconds not reached
        assert filt.should_emit("Soc", 73, 30.0) is False
        # max_seconds reached (61s > 60s) → force through despite small delta
        assert filt.should_emit("Soc", 73, 61.0) is True

    def test_max_seconds_zero_disables_staleness(self) -> None:
        """max_seconds=0 means staleness gate is inactive."""
        filters = {"Soc": FieldFilter(granularity=5.0, throttle_seconds=1.0, max_seconds=0.0)}
        filt = DualGateFilter(filters)
        assert filt.should_emit("Soc", 72, 0.0) is True
        filt.record_emit("Soc", 72, 0.0)
        # Even after a long time, small delta still blocked
        assert filt.should_emit("Soc", 73, 999.0) is False

    def test_max_seconds_respects_throttle(self) -> None:
        """Staleness gate doesn't bypass the throttle gate."""
        filters = {"Soc": FieldFilter(granularity=5.0, throttle_seconds=10.0, max_seconds=5.0)}
        filt = DualGateFilter(filters)
        assert filt.should_emit("Soc", 72, 0.0) is True
        filt.record_emit("Soc", 72, 0.0)
        # max_seconds elapsed (6s > 5s) but throttle blocks (6s < 10s)
        assert filt.should_emit("Soc", 73, 6.0) is False
        # Both staleness and throttle satisfied (11s > 10s > 5s)
        assert filt.should_emit("Soc", 73, 11.0) is True

    def test_max_seconds_on_first_value(self) -> None:
        """First value always emits regardless of max_seconds config."""
        filters = {"Soc": FieldFilter(granularity=5.0, throttle_seconds=1.0, max_seconds=60.0)}
        filt = DualGateFilter(filters)
        assert filt.should_emit("Soc", 72, 0.0) is True

    def test_max_seconds_resets_after_forced_emit(self) -> None:
        """After staleness forces an emit, timer resets normally."""
        filters = {"Soc": FieldFilter(granularity=5.0, throttle_seconds=1.0, max_seconds=10.0)}
        filt = DualGateFilter(filters)
        assert filt.should_emit("Soc", 72, 0.0) is True
        filt.record_emit("Soc", 72, 0.0)
        # Force emit at 11s via staleness
        assert filt.should_emit("Soc", 73, 11.0) is True
        filt.record_emit("Soc", 73, 11.0)
        # Small delta again, staleness timer reset — not stale yet
        assert filt.should_emit("Soc", 74, 15.0) is False
        # Stale again at 22s (11s since last emit at 11s)
        assert filt.should_emit("Soc", 74, 22.0) is True

    def test_max_seconds_with_location(self) -> None:
        """Staleness gate works for location fields too."""
        filters = {
            "Location": FieldFilter(granularity=50.0, throttle_seconds=1.0, max_seconds=30.0),
        }
        filt = DualGateFilter(filters)
        loc = {"latitude": 40.7128, "longitude": -74.0060}
        assert filt.should_emit("Location", loc, 0.0) is True
        filt.record_emit("Location", loc, 0.0)
        # Same location, not stale yet
        assert filt.should_emit("Location", loc, 15.0) is False
        # Same location, but stale (31s > 30s)
        assert filt.should_emit("Location", loc, 31.0) is True



"""Tests for EventEmitter telemetry-to-OpenClaw event mapping."""

from __future__ import annotations

from datetime import UTC, datetime

import pytest

from tescmd.openclaw.emitter import EventEmitter


@pytest.fixture()
def emitter() -> EventEmitter:
    return EventEmitter(client_id="test-bridge")


class TestLocationEvent:
    def test_valid_location(self, emitter: EventEmitter) -> None:
        event = emitter.to_event(
            "Location",
            {"latitude": 40.7128, "longitude": -74.006, "heading": 90.0, "speed": 30.0},
            vin="VIN1",
        )
        assert event is not None
        assert event["method"] == "req:agent"
        params = event["params"]
        assert params["event_type"] == "location"
        assert params["vin"] == "VIN1"
        assert params["data"]["latitude"] == 40.7128
        assert params["data"]["longitude"] == -74.006
        assert params["data"]["heading"] == 90.0
        assert params["data"]["speed"] == 30.0

    def test_location_missing_optional_fields(self, emitter: EventEmitter) -> None:
        event = emitter.to_event(
            "Location",
            {"latitude": 40.0, "longitude": -74.0},
            vin="VIN1",
        )
        assert event is not None
        assert event["params"]["data"]["heading"] == 0
        assert event["params"]["data"]["speed"] == 0

    def test_invalid_location_returns_none(self, emitter: EventEmitter) -> None:
        assert emitter.to_event("Location", "bad", vin="VIN1") is None
        assert emitter.to_event("Location", {}, vin="VIN1") is None


class TestBatteryEvents:
    def test_soc_event(self, emitter: EventEmitter) -> None:
        event = emitter.to_event("Soc", 72.5, vin="VIN1")
        assert event is not None
        assert event["params"]["event_type"] == "battery"
        assert event["params"]["data"]["battery_level"] == 72.5

    def test_battery_level_event(self, emitter: EventEmitter) -> None:
        event = emitter.to_event("BatteryLevel", 85, vin="VIN1")
        assert event is not None
        assert event["params"]["event_type"] == "battery"
        assert event["params"]["data"]["battery_level"] == 85.0

    def test_range_event(self, emitter: EventEmitter) -> None:
        event = emitter.to_event("EstBatteryRange", 250.5, vin="VIN1")
        assert event is not None
        assert event["params"]["event_type"] == "battery"
        assert event["params"]["data"]["range_miles"] == 250.5

    def test_invalid_soc_returns_none(self, emitter: EventEmitter) -> None:
        assert emitter.to_event("Soc", "bad", vin="VIN1") is None


class TestTempEvents:
    def test_inside_temp(self, emitter: EventEmitter) -> None:
        event = emitter.to_event("InsideTemp", 22.0, vin="VIN1")
        assert event is not None
        assert event["params"]["event_type"] == "inside_temp"
        # 22C = 71.6F
        assert event["params"]["data"]["inside_temp_f"] == 71.6

    def test_outside_temp(self, emitter: EventEmitter) -> None:
        event = emitter.to_event("OutsideTemp", 0.0, vin="VIN1")
        assert event is not None
        assert event["params"]["event_type"] == "outside_temp"
        # 0C = 32F
        assert event["params"]["data"]["outside_temp_f"] == 32.0

    def test_invalid_temp_returns_none(self, emitter: EventEmitter) -> None:
        assert emitter.to_event("InsideTemp", "hot", vin="VIN1") is None


class TestSpeedEvent:
    def test_speed(self, emitter: EventEmitter) -> None:
        event = emitter.to_event("VehicleSpeed", 65.0, vin="VIN1")
        assert event is not None
        assert event["params"]["event_type"] == "speed"
        assert event["params"]["data"]["speed_mph"] == 65.0


class TestChargeStateEvents:
    def test_charging_state(self, emitter: EventEmitter) -> None:
        event = emitter.to_event("ChargeState", "Charging", vin="VIN1")
        assert event is not None
        assert event["params"]["event_type"] == "charge_started"
        assert event["params"]["data"]["state"] == "Charging"

    def test_complete_state(self, emitter: EventEmitter) -> None:
        event = emitter.to_event("ChargeState", "Complete", vin="VIN1")
        assert event is not None
        assert event["params"]["event_type"] == "charge_complete"

    def test_stopped_state(self, emitter: EventEmitter) -> None:
        event = emitter.to_event("ChargeState", "Stopped", vin="VIN1")
        assert event is not None
        assert event["params"]["event_type"] == "charge_stopped"

    def test_disconnected_state(self, emitter: EventEmitter) -> None:
        event = emitter.to_event("ChargeState", "Disconnected", vin="VIN1")
        assert event is not None
        assert event["params"]["event_type"] == "charge_stopped"

    def test_detailed_charge_state(self, emitter: EventEmitter) -> None:
        event = emitter.to_event("DetailedChargeState", "Starting", vin="VIN1")
        assert event is not None
        assert event["params"]["event_type"] == "charge_started"


class TestSecurityEvents:
    def test_locked(self, emitter: EventEmitter) -> None:
        event = emitter.to_event("Locked", True, vin="VIN1")
        assert event is not None
        assert event["params"]["event_type"] == "security_changed"
        assert event["params"]["data"]["field"] == "locked"
        assert event["params"]["data"]["value"] is True

    def test_sentry_mode(self, emitter: EventEmitter) -> None:
        event = emitter.to_event("SentryMode", False, vin="VIN1")
        assert event is not None
        assert event["params"]["event_type"] == "security_changed"
        assert event["params"]["data"]["field"] == "sentrymode"


class TestGearEvent:
    def test_gear(self, emitter: EventEmitter) -> None:
        event = emitter.to_event("Gear", "D", vin="VIN1")
        assert event is not None
        assert event["params"]["event_type"] == "gear_changed"
        assert event["params"]["data"]["gear"] == "D"


class TestGenericFallback:
    def test_unknown_field_produces_generic_event(self, emitter: EventEmitter) -> None:
        event = emitter.to_event("PackVoltage", 398.2, vin="VIN1")
        assert event is not None
        assert event["params"]["event_type"] == "pack_voltage"
        data = event["params"]["data"]
        assert data["field"] == "PackVoltage"
        assert data["value"] == 398.2

    def test_generic_event_snake_case_naming(self, emitter: EventEmitter) -> None:
        cases = {
            "HvacFanSpeed": "hvac_fan_speed",
            "TpmsPressureFl": "tpms_pressure_fl",
            "ChargeLimitSoc": "charge_limit_soc",
            "Odometer": "odometer",
            "BMSState": "bms_state",
        }
        for field_name, expected_type in cases.items():
            event = emitter.to_event(field_name, 1, vin="VIN1")
            assert event is not None, f"No event for {field_name}"
            assert event["params"]["event_type"] == expected_type, (
                f"{field_name} → {event['params']['event_type']}, expected {expected_type}"
            )

    def test_generic_event_with_dict_value(self, emitter: EventEmitter) -> None:
        val = {"a": 1, "b": "two"}
        event = emitter.to_event("CustomSensor", val, vin="VIN1")
        assert event is not None
        data = event["params"]["data"]
        assert data["value"] == val

    def test_generic_event_preserves_envelope(self, emitter: EventEmitter) -> None:
        event = emitter.to_event("PackVoltage", 398.2, vin="VIN1")
        assert event is not None
        assert event["method"] == "req:agent"
        assert event["params"]["source"] == "test-bridge"
        assert event["params"]["vin"] == "VIN1"


class TestTimestamp:
    def test_custom_timestamp(self, emitter: EventEmitter) -> None:
        ts = datetime(2026, 1, 31, 12, 0, 0, tzinfo=UTC)
        event = emitter.to_event("Soc", 72, vin="VIN1", timestamp=ts)
        assert event is not None
        assert event["params"]["timestamp"] == "2026-01-31T12:00:00+00:00"

    def test_source_matches_client_id(self, emitter: EventEmitter) -> None:
        event = emitter.to_event("Soc", 72, vin="VIN1")
        assert event is not None
        assert event["params"]["source"] == "test-bridge"



"""Vehicle command API — wraps POST /api/1/vehicles/{vin}/command/{name}."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from tescmd.models.command import CommandResponse

if TYPE_CHECKING:
    from tescmd.api.client import TeslaFleetClient


class CommandAPI:
    """Vehicle command operations (composition over :class:`TeslaFleetClient`)."""

    def __init__(self, client: TeslaFleetClient) -> None:
        self._client = client

    async def _command(
        self, vin: str, command: str, body: dict[str, Any] | None = None
    ) -> CommandResponse:
        path = f"/api/1/vehicles/{vin}/command/{command}"
        kwargs: dict[str, Any] = {}
        if body is not None:
            kwargs["json"] = body
        data = await self._client.post(path, **kwargs)
        return CommandResponse.model_validate(data)

    # ------------------------------------------------------------------
    # Key enrollment
    # ------------------------------------------------------------------
    # Note: Initial key enrollment is NOT available via REST or signed_command.
    # The Tesla Go SDK explicitly blocks add_key_request for Fleet API with
    # ErrRequiresBLE. For Fleet API apps, enrollment happens through the
    # Tesla app portal flow: https://tesla.com/_ak/<domain>
    # See cli/key.py for the enrollment flow implementation.

    # ------------------------------------------------------------------
    # Security / convenience commands
    # ------------------------------------------------------------------

    async def auto_secure_vehicle(self, vin: str) -> CommandResponse:
        """Close falcon-wing doors and lock (Model X only)."""
        return await self._command(vin, "auto_secure_vehicle")

    # ------------------------------------------------------------------
    # Charging commands
    # ------------------------------------------------------------------

    async def charge_start(self, vin: str) -> CommandResponse:
        return await self._command(vin, "charge_start")

    async def charge_stop(self, vin: str) -> CommandResponse:
        return await self._command(vin, "charge_stop")

    async def charge_standard(self, vin: str) -> CommandResponse:
        return await self._command(vin, "charge_standard")

    async def charge_max_range(self, vin: str) -> CommandResponse:
        return await self._command(vin, "charge_max_range")

    async def charge_port_door_open(self, vin: str) -> CommandResponse:
        return await self._command(vin, "charge_port_door_open")

    async def charge_port_door_close(self, vin: str) -> CommandResponse:
        return await self._command(vin, "charge_port_door_close")

    async def set_charge_limit(self, vin: str, *, percent: int) -> CommandResponse:
        return await self._command(vin, "set_charge_limit", {"percent": percent})

    async def set_charging_amps(self, vin: str, *, charging_amps: int) -> CommandResponse:
        return await self._command(vin, "set_charging_amps", {"charging_amps": charging_amps})

    async def set_scheduled_charging(
        self, vin: str, *, enable: bool, time: int
    ) -> CommandResponse:
        return await self._command(vin, "set_scheduled_charging", {"enable": enable, "time": time})

    async def set_scheduled_departure(
        self,
        vin: str,
        *,
        enable: bool,
        departure_time: int,
        preconditioning_enabled: bool = False,
        preconditioning_weekdays_only: bool = False,
        off_peak_charging_enabled: bool = False,
        off_peak_charging_weekdays_only: bool = False,
        end_off_peak_time: int = 0,
    ) -> CommandResponse:
        return await self._command(
            vin,
            "set_scheduled_departure",
            {
                "enable": enable,
                "departure_time": departure_time,
                "preconditioning_enabled": preconditioning_enabled,
                "preconditioning_weekdays_only": preconditioning_weekdays_only,
                "off_peak_charging_enabled": off_peak_charging_enabled,
                "off_peak_charging_weekdays_only": off_peak_charging_weekdays_only,
                "end_off_peak_time": end_off_peak_time,
            },
        )

    async def add_precondition_schedule(
        self,
        vin: str,
        *,
        lat: float | None = None,
        lon: float | None = None,
        precondition_time: int | None = None,
        one_time: bool | None = None,
        days_of_week: str | None = None,
        id: int | None = None,
        enabled: bool | None = None,
    ) -> CommandResponse:
        body: dict[str, Any] = {}
        if lat is not None:
            body["lat"] = lat
        if lon is not None:
            body["lon"] = lon
        if precondition_time is not None:
            body["precondition_time"] = precondition_time
        if one_time is not None:
            body["one_time"] = one_time
        if days_of_week is not None:
            body["days_of_week"] = days_of_week
        if id is not None:
            body["id"] = id
        if enabled is not None:
            body["enabled"] = enabled
        return await self._command(vin, "add_precondition_schedule", body)

    async def remove_precondition_schedule(self, vin: str, *, id: int) -> CommandResponse:
        return await self._command(vin, "remove_precondition_schedule", {"id": id})

    async def batch_remove_precondition_schedules(
        self, vin: str, *, home: bool, work: bool, other: bool
    ) -> CommandResponse:
        """Remove precondition schedules by location type (home/work/other)."""
        return await self._command(
            vin,
            "batch_remove_precondition_schedules",
            {"home": home, "work": work, "other": other},
        )

    # ------------------------------------------------------------------
    # Climate commands
    # ------------------------------------------------------------------

    async def auto_conditioning_start(self, vin: str) -> CommandResponse:
        return await self._command(vin, "auto_conditioning_start")

    async def auto_conditioning_stop(self, vin: str) -> CommandResponse:
        return await self._command(vin, "auto_conditioning_stop")

    async def set_temps(
        self, vin: str, *, driver_temp: float, passenger_temp: float
    ) -> CommandResponse:
        return await self._command(
            vin,
            "set_temps",
            {"driver_temp": driver_temp, "passenger_temp": passenger_temp},
        )

    async def set_preconditioning_max(
        self, vin: str, *, on: bool, manual_override: bool = False
    ) -> CommandResponse:
        return await self._command(
            vin, "set_preconditioning_max", {"on": on, "manual_override": manual_override}
        )

    async def remote_seat_heater_request(
        self, vin: str, *, seat_position: int, level: int
    ) -> CommandResponse:
        return await self._command(
            vin, "remote_seat_heater_request", {"seat_position": seat_position, "level": level}
        )

    async def remote_seat_cooler_request(
        self, vin: str, *, seat_position: int, seat_cooler_level: int
    ) -> CommandResponse:
        return await self._command(
            vin,
            "remote_seat_cooler_request",
            {"seat_position": seat_position, "seat_cooler_level": seat_cooler_level},
        )

    async def remote_steering_wheel_heater_request(self, vin: str, *, on: bool) -> CommandResponse:
        return await self._command(vin, "remote_steering_wheel_heater_request", {"on": on})

    async def set_cabin_overheat_protection(
        self, vin: str, *, on: bool, fan_only: bool = False
    ) -> CommandResponse:
        return await self._command(
            vin,
            "set_cabin_overheat_protection",
            {"on": on, "fan_only": fan_only},
        )

    async def set_climate_keeper_mode(
        self, vin: str, *, climate_keeper_mode: int, manual_override: bool = False
    ) -> CommandResponse:
        return await self._command(
            vin,
            "set_climate_keeper_mode",
            {"climate_keeper_mode": climate_keeper_mode, "manual_override": manual_override},
        )

    async def set_cop_temp(self, vin: str, *, cop_temp: int) -> CommandResponse:
        """Set cabin overheat protection temperature (0=low, 1=medium, 2=high)."""
        return await self._command(vin, "set_cop_temp", {"cop_temp": cop_temp})

    async def remote_auto_seat_climate_request(
        self, vin: str, *, auto_seat_position: int, auto_climate_on: bool
    ) -> CommandResponse:
        return await self._command(
            vin,
            "remote_auto_seat_climate_request",
            {"auto_seat_position": auto_seat_position, "auto_climate_on": auto_climate_on},
        )

    async def remote_auto_steering_wheel_heat_climate_request(
        self, vin: str, *, on: bool
    ) -> CommandResponse:
        return await self._command(
            vin, "remote_auto_steering_wheel_heat_climate_request", {"on": on}
        )

    async def remote_steering_wheel_heat_level_request(
        self, vin: str, *, level: int
    ) -> CommandResponse:
        return await self._command(
            vin, "remote_steering_wheel_heat_level_request", {"level": level}
        )

    # ------------------------------------------------------------------
    # Security commands
    # ------------------------------------------------------------------

    async def door_lock(self, vin: str) -> CommandResponse:
        return await self._command(vin, "door_lock")

    async def door_unlock(self, vin: str) -> CommandResponse:
        return await self._command(vin, "door_unlock")

    async def set_sentry_mode(self, vin: str, *, on: bool) -> CommandResponse:
        return await self._command(vin, "set_sentry_mode", {"on": on})

    async def set_valet_mode(
        self, vin: str, *, on: bool, password: str | None = None
    ) -> CommandResponse:
        body: dict[str, Any] = {"on": on}
        if password is not None:
            body["password"] = password
        return await self._command(vin, "set_valet_mode", body)

    async def reset_valet_pin(self, vin: str) -> CommandResponse:
        return await self._command(vin, "reset_valet_pin")

    async def speed_limit_activate(self, vin: str, *, pin: str) -> CommandResponse:
        return await self._command(vin, "speed_limit_activate", {"pin": pin})

    async def speed_limit_deactivate(self, vin: str, *, pin: str) -> CommandResponse:
        return await self._command(vin, "speed_limit_deactivate", {"pin": pin})

    async def speed_limit_set_limit(self, vin: str, *, limit_mph: float) -> CommandResponse:
        return await self._command(vin, "speed_limit_set_limit", {"limit_mph": limit_mph})

    async def reset_pin_to_drive_pin(self, vin: str) -> CommandResponse:
        return await self._command(vin, "reset_pin_to_drive_pin")

    async def clear_pin_to_drive_admin(self, vin: str) -> CommandResponse:
        return await self._command(vin, "clear_pin_to_drive_admin")

    async def speed_limit_clear_pin(self, vin: str, *, pin: str) -> CommandResponse:
        return await self._command(vin, "speed_limit_clear_pin", {"pin": pin})

    async def speed_limit_clear_pin_admin(self, vin: str) -> CommandResponse:
        return await self._command(vin, "speed_limit_clear_pin_admin")

    async def remote_start_drive(self, vin: str) -> CommandResponse:
        return await self._command(vin, "remote_start_drive")

    async def flash_lights(self, vin: str) -> CommandResponse:
        return await self._command(vin, "flash_lights")

    async def honk_horn(self, vin: str) -> CommandResponse:
        return await self._command(vin, "honk_horn")

    # ------------------------------------------------------------------
    # Media commands
    # ------------------------------------------------------------------

    async def media_toggle_playback(self, vin: str) -> CommandResponse:
        return await self._command(vin, "media_toggle_playback")

    async def media_next_track(self, vin: str) -> CommandResponse:
        return await self._command(vin, "media_next_track")

    async def media_prev_track(self, vin: str) -> CommandResponse:
        return await self._command(vin, "media_prev_track")

    async def media_next_fav(self, vin: str) -> CommandResponse:
        return await self._command(vin, "media_next_fav")

    async def media_prev_fav(self, vin: str) -> CommandResponse:
        return await self._command(vin, "media_prev_fav")

    async def media_volume_up(self, vin: str) -> CommandResponse:
        return await self._command(vin, "media_volume_up")

    async def media_volume_down(self, vin: str) -> CommandResponse:
        return await self._command(vin, "media_volume_down")

    async def adjust_volume(self, vin: str, *, volume: float) -> CommandResponse:
        return await self._command(vin, "adjust_volume", {"volume": volume})

    # ------------------------------------------------------------------
    # Navigation commands
    # ------------------------------------------------------------------

    async def share(self, vin: str, *, address: str) -> CommandResponse:
        import time as _time

        body = {
            "type": "share_ext_content_raw",
            "value": {"android.intent.extra.TEXT": address},
            "locale": "en-US",
            "timestamp_ms": str(int(_time.time() * 1000)),
        }
        path = f"/api/1/vehicles/{vin}/command/share"
        data = await self._client.post(path, json=body)
        return CommandResponse.model_validate(data)

    async def navigation_gps_request(
        self, vin: str, *, lat: float, lon: float, order: int | None = None
    ) -> CommandResponse:
        body: dict[str, Any] = {"lat": lat, "lon": lon}
        if order is not None:
            body["order"] = order
        return await self._command(vin, "navigation_gps_request", body)

    async def navigation_sc_request(
        self, vin: str, *, id: int = 0, order: int = 0
    ) -> CommandResponse:
        return await self._command(vin, "navigation_sc_request", {"id": id, "order": order})

    async def trigger_homelink(self, vin: str, *, lat: float, lon: float) -> CommandResponse:
        return await self._command(vin, "trigger_homelink", {"lat": lat, "lon": lon})

    async def navigation_request(
        self,
        vin: str,
        *,
        type: str = "share_ext_content_raw",
        locale: str = "en-US",
        timestamp_ms: str = "",
        value: dict[str, Any] | None = None,
    ) -> CommandResponse:
        """Legacy navigation request (REST-only, deprecated in favour of 'share')."""
        import time as _time

        body: dict[str, Any] = {
            "type": type,
            "locale": locale,
            "timestamp_ms": timestamp_ms or str(int(_time.time() * 1000)),
        }
        if value is not None:
            body["value"] = value
        path = f"/api/1/vehicles/{vin}/command/navigation_request"
        data = await self._client.post(path, json=body)
        return CommandResponse.model_validate(data)

    async def navigation_waypoints_request(self, vin: str, *, waypoints: str) -> CommandResponse:
        return await self._command(vin, "navigation_waypoints_request", {"waypoints": waypoints})

    # ------------------------------------------------------------------
    # Software commands
    # ------------------------------------------------------------------

    async def schedule_software_update(self, vin: str, *, offset_sec: int) -> CommandResponse:
        return await self._command(vin, "schedule_software_update", {"offset_sec": offset_sec})

    async def cancel_software_update(self, vin: str) -> CommandResponse:
        return await self._command(vin, "cancel_software_update")

    # ------------------------------------------------------------------
    # Bioweapon / sunroof / PIN-to-drive / guest-mode / erase / boombox
    # ------------------------------------------------------------------

    async def set_bioweapon_mode(
        self, vin: str, *, on: bool, manual_override: bool = False
    ) -> CommandResponse:
        return await self._command(
            vin, "set_bioweapon_mode", {"on": on, "manual_override": manual_override}
        )

    async def sun_roof_control(self, vin: str, *, state: str) -> CommandResponse:
        return await self._command(vin, "sun_roof_control", {"state": state})

    async def set_pin_to_drive(
        self, vin: str, *, on: bool, password: str | None = None
    ) -> CommandResponse:
        body: dict[str, Any] = {"on": on}
        if password is not None:
            body["password"] = password
        return await self._command(vin, "set_pin_to_drive", body)

    async def guest_mode(self, vin: str, *, enable: bool) -> CommandResponse:
        return await self._command(vin, "guest_mode", {"enable": enable})

    async def erase_user_data(self, vin: str) -> CommandResponse:
        return await self._command(vin, "erase_user_data")

    async def remote_boombox(self, vin: str, *, sound: int = 2000) -> CommandResponse:
        return await self._command(vin, "remote_boombox", {"sound": sound})

    # ------------------------------------------------------------------
    # Charge schedule commands (firmware 2024.26+)
    # ------------------------------------------------------------------

    async def add_charge_schedule(
        self,
        vin: str,
        *,
        lat: float | None = None,
        lon: float | None = None,
        start_time: int | None = None,
        start_enabled: bool | None = None,
        end_time: int | None = None,
        end_enabled: bool | None = None,
        days_of_week: str | None = None,
        id: int | None = None,
        enabled: bool | None = None,
        one_time: bool | None = None,
    ) -> CommandResponse:
        body: dict[str, Any] = {}
        if lat is not None:
            body["lat"] = lat
        if lon is not None:
            body["lon"] = lon
        if start_time is not None:
            body["start_time"] = start_time
        if start_enabled is not None:
            body["start_enabled"] = start_enabled
        if end_time is not None:
            body["end_time"] = end_time
        if end_enabled is not None:
            body["end_enabled"] = end_enabled
        if days_of_week is not None:
            body["days_of_week"] = days_of_week
        if id is not None:
            body["id"] = id
        if enabled is not None:
            body["enabled"] = enabled
        if one_time is not None:
            body["one_time"] = one_time
        return await self._command(vin, "add_charge_schedule", body)

    async def remove_charge_schedule(self, vin: str, *, id: int) -> CommandResponse:
        return await self._command(vin, "remove_charge_schedule", {"id": id})

    async def batch_remove_charge_schedules(
        self, vin: str, *, home: bool, work: bool, other: bool
    ) -> CommandResponse:
        """Remove charge schedules by location type (home/work/other)."""
        return await self._command(
            vin,
            "batch_remove_charge_schedules",
            {"home": home, "work": work, "other": other},
        )

    # ------------------------------------------------------------------
    # Vehicle name / calendar
    # ------------------------------------------------------------------

    async def set_vehicle_name(self, vin: str, *, vehicle_name: str) -> CommandResponse:
        return await self._command(vin, "set_vehicle_name", {"vehicle_name": vehicle_name})

    async def upcoming_calendar_entries(self, vin: str, *, calendar_data: str) -> CommandResponse:
        return await self._command(
            vin, "upcoming_calendar_entries", {"calendar_data": calendar_data}
        )

    # ------------------------------------------------------------------
    # Trunk / window commands
    # ------------------------------------------------------------------

    async def actuate_trunk(self, vin: str, *, which_trunk: str) -> CommandResponse:
        return await self._command(vin, "actuate_trunk", {"which_trunk": which_trunk})

    async def window_control(
        self, vin: str, *, command: str, lat: float = 0.0, lon: float = 0.0
    ) -> CommandResponse:
        return await self._command(
            vin, "window_control", {"command": command, "lat": lat, "lon": lon}
        )

    # ------------------------------------------------------------------
    # Tonneau cover commands (Cybertruck)
    # ------------------------------------------------------------------

    async def open_tonneau(self, vin: str) -> CommandResponse:
        return await self._command(vin, "open_tonneau")

    async def close_tonneau(self, vin: str) -> CommandResponse:
        return await self._command(vin, "close_tonneau")

    async def stop_tonneau(self, vin: str) -> CommandResponse:
        return await self._command(vin, "stop_tonneau")

    # ------------------------------------------------------------------
    # Power management commands
    # ------------------------------------------------------------------

    async def set_low_power_mode(self, vin: str, *, enable: bool) -> CommandResponse:
        return await self._command(vin, "set_low_power_mode", {"enable": enable})

    async def keep_accessory_power_mode(self, vin: str, *, enable: bool) -> CommandResponse:
        return await self._command(vin, "keep_accessory_power_mode", {"enable": enable})

    # ------------------------------------------------------------------
    # Managed charging (fleet)
    # ------------------------------------------------------------------

    async def set_managed_charge_current_request(
        self, vin: str, *, charging_amps: int
    ) -> CommandResponse:
        return await self._command(
            vin, "set_managed_charge_current_request", {"charging_amps": charging_amps}
        )

    async def set_managed_charger_location(
        self, vin: str, *, location: dict[str, Any]
    ) -> CommandResponse:
        return await self._command(vin, "set_managed_charger_location", location)

    async def set_managed_scheduled_charging_time(self, vin: str, *, time: int) -> CommandResponse:
        return await self._command(vin, "set_managed_scheduled_charging_time", {"time": time})



"""Tests for EC key generation and loading."""

from __future__ import annotations

import stat
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from pathlib import Path
from cryptography.hazmat.primitives.asymmetric import ec

from tescmd.crypto.keys import (
    generate_ec_key_pair,
    get_key_fingerprint,
    get_public_key_path,
    has_key_pair,
    load_private_key,
    load_public_key_pem,
)


class TestGenerateEcKeyPair:
    def test_creates_key_files(self, tmp_path: Path) -> None:
        priv, pub = generate_ec_key_pair(tmp_path)
        assert priv.exists()
        assert pub.exists()
        assert priv.name == "private_key.pem"
        assert pub.name == "public_key.pem"

    def test_private_key_permissions(self, tmp_path: Path) -> None:
        priv, _pub = generate_ec_key_pair(tmp_path)
        mode = priv.stat().st_mode
        assert mode & stat.S_IRUSR  # owner can read
        assert mode & stat.S_IWUSR  # owner can write
        assert not (mode & stat.S_IRGRP)  # group cannot read
        assert not (mode & stat.S_IROTH)  # others cannot read

    def test_refuses_overwrite_by_default(self, tmp_path: Path) -> None:
        generate_ec_key_pair(tmp_path)
        with pytest.raises(FileExistsError):
            generate_ec_key_pair(tmp_path)

    def test_overwrite_flag(self, tmp_path: Path) -> None:
        priv1, _ = generate_ec_key_pair(tmp_path)
        content1 = priv1.read_bytes()
        priv2, _ = generate_ec_key_pair(tmp_path, overwrite=True)
        content2 = priv2.read_bytes()
        assert content1 != content2  # new key generated

    def test_creates_intermediate_directories(self, tmp_path: Path) -> None:
        key_dir = tmp_path / "deep" / "nested" / "keys"
        priv, pub = generate_ec_key_pair(key_dir)
        assert priv.exists()
        assert pub.exists()

    def test_pem_format(self, tmp_path: Path) -> None:
        priv, pub = generate_ec_key_pair(tmp_path)
        priv_text = priv.read_text()
        pub_text = pub.read_text()
        assert priv_text.startswith("-----BEGIN PRIVATE KEY-----")
        assert pub_text.startswith("-----BEGIN PUBLIC KEY-----")


class TestLoadPrivateKey:
    def test_loads_valid_key(self, tmp_path: Path) -> None:
        generate_ec_key_pair(tmp_path)
        key = load_private_key(tmp_path)
        assert isinstance(key, ec.EllipticCurvePrivateKey)

    def test_raises_on_missing_key(self, tmp_path: Path) -> None:
        with pytest.raises(FileNotFoundError):
            load_private_key(tmp_path)


class TestLoadPublicKeyPem:
    def test_loads_pem_string(self, tmp_path: Path) -> None:
        generate_ec_key_pair(tmp_path)
        pem = load_public_key_pem(tmp_path)
        assert isinstance(pem, str)
        assert "BEGIN PUBLIC KEY" in pem

    def test_raises_on_missing_key(self, tmp_path: Path) -> None:
        with pytest.raises(FileNotFoundError):
            load_public_key_pem(tmp_path)


class TestGetPublicKeyPath:
    def test_returns_path(self, tmp_path: Path) -> None:
        path = get_public_key_path(tmp_path)
        assert path == tmp_path / "public_key.pem"


class TestHasKeyPair:
    def test_false_when_empty(self, tmp_path: Path) -> None:
        assert has_key_pair(tmp_path) is False

    def test_true_after_generation(self, tmp_path: Path) -> None:
        generate_ec_key_pair(tmp_path)
        assert has_key_pair(tmp_path) is True

    def test_false_with_only_private(self, tmp_path: Path) -> None:
        generate_ec_key_pair(tmp_path)
        (tmp_path / "public_key.pem").unlink()
        assert has_key_pair(tmp_path) is False


class TestGetKeyFingerprint:
    def test_returns_hex_string(self, tmp_path: Path) -> None:
        generate_ec_key_pair(tmp_path)
        fp = get_key_fingerprint(tmp_path)
        assert isinstance(fp, str)
        assert len(fp) == 64  # SHA-256 hex = 64 chars
        int(fp, 16)  # valid hex

    def test_deterministic(self, tmp_path: Path) -> None:
        generate_ec_key_pair(tmp_path)
        fp1 = get_key_fingerprint(tmp_path)
        fp2 = get_key_fingerprint(tmp_path)
        assert fp1 == fp2

    def test_different_keys_different_fingerprints(self, tmp_path: Path) -> None:
        dir1 = tmp_path / "a"
        dir2 = tmp_path / "b"
        generate_ec_key_pair(dir1)
        generate_ec_key_pair(dir2)
        assert get_key_fingerprint(dir1) != get_key_fingerprint(dir2)

    def test_raises_on_missing_key(self, tmp_path: Path) -> None:
        with pytest.raises(FileNotFoundError):
            get_key_fingerprint(tmp_path)



"""High-level Vehicle API built on top of TeslaFleetClient."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from tescmd.models.sharing import ShareDriverInfo
from tescmd.models.vehicle import NearbyChargingSites, Vehicle, VehicleData

if TYPE_CHECKING:
    from tescmd.api.client import TeslaFleetClient


class VehicleAPI:
    """Vehicle-related API operations (composition over TeslaFleetClient)."""

    def __init__(self, client: TeslaFleetClient) -> None:
        self._client = client

    async def list_vehicles(self) -> list[Vehicle]:
        """Return all vehicles associated with the account."""
        data = await self._client.get("/api/1/vehicles")
        raw_list: list[dict[str, object]] = data.get("response", [])
        return [Vehicle.model_validate(v) for v in raw_list]

    async def get_vehicle(self, vin: str) -> Vehicle:
        """Fetch a single vehicle by VIN."""
        data = await self._client.get(f"/api/1/vehicles/{vin}")
        return Vehicle.model_validate(data["response"])

    async def get_vehicle_data(
        self,
        vin: str,
        *,
        endpoints: list[str] | None = None,
    ) -> VehicleData:
        """Fetch full vehicle data, optionally filtered to *endpoints*."""
        path = f"/api/1/vehicles/{vin}/vehicle_data"
        params: dict[str, str] = {}
        if endpoints:
            params["endpoints"] = ";".join(endpoints)
        data = await self._client.get(path, params=params)
        return VehicleData.model_validate(data["response"])

    async def wake(self, vin: str) -> Vehicle:
        """Send a wake-up command and return the vehicle state."""
        data = await self._client.post(f"/api/1/vehicles/{vin}/wake_up")
        return Vehicle.model_validate(data["response"])

    async def mobile_enabled(self, vin: str) -> bool:
        """Check if mobile access is enabled for the vehicle."""
        data = await self._client.get(f"/api/1/vehicles/{vin}/mobile_enabled")
        return bool(data.get("response", False))

    async def nearby_charging_sites(self, vin: str) -> NearbyChargingSites:
        """Fetch nearby Superchargers and destination chargers."""
        data = await self._client.get(f"/api/1/vehicles/{vin}/nearby_charging_sites")
        return NearbyChargingSites.model_validate(data.get("response", {}))

    async def recent_alerts(self, vin: str) -> list[dict[str, Any]]:
        """Fetch recent vehicle alerts."""
        data = await self._client.get(f"/api/1/vehicles/{vin}/recent_alerts")
        result: list[dict[str, Any]] = data.get("response", [])
        return result

    async def release_notes(self, vin: str) -> dict[str, Any]:
        """Fetch firmware release notes."""
        data = await self._client.get(f"/api/1/vehicles/{vin}/release_notes")
        result: dict[str, Any] = data.get("response", {})
        return result

    async def service_data(self, vin: str) -> dict[str, Any]:
        """Fetch vehicle service data."""
        data = await self._client.get(f"/api/1/vehicles/{vin}/service_data")
        result: dict[str, Any] = data.get("response", {})
        return result

    async def list_drivers(self, vin: str) -> list[ShareDriverInfo]:
        """List drivers associated with the vehicle."""
        data = await self._client.get(f"/api/1/vehicles/{vin}/drivers")
        raw_list: list[dict[str, Any]] = data.get("response", [])
        return [ShareDriverInfo.model_validate(d) for d in raw_list]

    # ------------------------------------------------------------------
    # Extended vehicle data endpoints
    # ------------------------------------------------------------------

    async def eligible_subscriptions(self, vin: str) -> dict[str, Any]:
        """Check subscription eligibility for the vehicle."""
        data = await self._client.get(
            "/api/1/dx/vehicles/subscriptions/eligibility", params={"vin": vin}
        )
        result: dict[str, Any] = data.get("response", {})
        return result

    async def eligible_upgrades(self, vin: str) -> dict[str, Any]:
        """Check upgrade eligibility for the vehicle."""
        data = await self._client.get(
            "/api/1/dx/vehicles/upgrades/eligibility", params={"vin": vin}
        )
        result: dict[str, Any] = data.get("response", {})
        return result

    async def options(self, vin: str) -> dict[str, Any]:
        """Fetch vehicle option codes."""
        data = await self._client.get("/api/1/dx/vehicles/options", params={"vin": vin})
        result: dict[str, Any] = data.get("response", {})
        return result

    async def specs(self, vin: str) -> dict[str, Any]:
        """Fetch vehicle specifications."""
        data = await self._client.get(f"/api/1/vehicles/{vin}/specs")
        result: dict[str, Any] = data.get("response", {})
        return result

    async def warranty_details(self) -> dict[str, Any]:
        """Fetch warranty details."""
        data = await self._client.get("/api/1/dx/warranty/details")
        result: dict[str, Any] = data.get("response", {})
        return result

    async def fleet_status(self) -> dict[str, Any]:
        """Fetch fleet status for all vehicles."""
        data = await self._client.post("/api/1/vehicles/fleet_status")
        result: dict[str, Any] = data.get("response", {})
        return result

    async def fleet_telemetry_config(self, vin: str) -> dict[str, Any]:
        """Fetch fleet telemetry configuration for a vehicle."""
        data = await self._client.get(f"/api/1/vehicles/{vin}/fleet_telemetry_config")
        result: dict[str, Any] = data.get("response", {})
        return result

    async def fleet_telemetry_config_create(self, *, config: dict[str, Any]) -> dict[str, Any]:
        """Create or update fleet telemetry server configuration.

        .. deprecated::
            Tesla now requires the Vehicle Command HTTP Proxy for this
            endpoint.  Use :meth:`fleet_telemetry_config_create_jws` instead.
        """
        data = await self._client.post("/api/1/vehicles/fleet_telemetry_config", json=config)
        result: dict[str, Any] = data.get("response", {})
        return result

    async def fleet_telemetry_config_create_jws(
        self, *, vins: list[str], token: str
    ) -> dict[str, Any]:
        """Create fleet telemetry config via the JWS-signed endpoint.

        *token* is a compact JWS signed with the Tesla.SS256 algorithm
        (Schnorr/P-256).  See :func:`tescmd.crypto.schnorr.sign_fleet_telemetry_config`.
        """
        payload = {"vins": vins, "token": token}
        data = await self._client.post("/api/1/vehicles/fleet_telemetry_config_jws", json=payload)
        result: dict[str, Any] = data.get("response", {})
        return result

    async def fleet_telemetry_config_delete(self, vin: str) -> dict[str, Any]:
        """Remove fleet telemetry configuration from a vehicle."""
        data = await self._client.delete(f"/api/1/vehicles/{vin}/fleet_telemetry_config")
        result: dict[str, Any] = data.get("response", {})
        return result

    async def fleet_telemetry_errors(self, vin: str) -> dict[str, Any]:
        """Fetch fleet telemetry errors."""
        data = await self._client.get(f"/api/1/vehicles/{vin}/fleet_telemetry_errors")
        result: dict[str, Any] = data.get("response", {})
        return result



"""Tests for the FrameFanout telemetry dispatcher."""

from __future__ import annotations

from datetime import UTC, datetime
from unittest.mock import AsyncMock

import pytest

from tescmd.telemetry.decoder import TelemetryDatum, TelemetryFrame
from tescmd.telemetry.fanout import FrameFanout


def _make_frame(
    vin: str = "5YJ3E1EA0KF000001", field: str = "Soc", value: int = 80
) -> TelemetryFrame:
    return TelemetryFrame(
        vin=vin,
        created_at=datetime.now(UTC),
        data=[TelemetryDatum(field_name=field, field_id=3, value=value, value_type="int")],
    )


class TestFrameFanout:
    def test_empty_fanout(self) -> None:
        fanout = FrameFanout()
        assert fanout.sink_count == 0
        assert not fanout.has_sinks()

    def test_add_sink(self) -> None:
        fanout = FrameFanout()
        fanout.add_sink(AsyncMock())
        assert fanout.sink_count == 1
        assert fanout.has_sinks()

    def test_add_multiple_sinks(self) -> None:
        fanout = FrameFanout()
        fanout.add_sink(AsyncMock())
        fanout.add_sink(AsyncMock())
        fanout.add_sink(AsyncMock())
        assert fanout.sink_count == 3

    @pytest.mark.asyncio
    async def test_dispatches_to_single_sink(self) -> None:
        fanout = FrameFanout()
        sink = AsyncMock()
        fanout.add_sink(sink)

        frame = _make_frame()
        await fanout.on_frame(frame)

        sink.assert_awaited_once_with(frame)

    @pytest.mark.asyncio
    async def test_dispatches_to_all_sinks(self) -> None:
        fanout = FrameFanout()
        sink_a = AsyncMock()
        sink_b = AsyncMock()
        sink_c = AsyncMock()
        fanout.add_sink(sink_a)
        fanout.add_sink(sink_b)
        fanout.add_sink(sink_c)

        frame = _make_frame()
        await fanout.on_frame(frame)

        sink_a.assert_awaited_once_with(frame)
        sink_b.assert_awaited_once_with(frame)
        sink_c.assert_awaited_once_with(frame)

    @pytest.mark.asyncio
    async def test_failing_sink_does_not_stop_others(self) -> None:
        fanout = FrameFanout()
        sink_a = AsyncMock()
        sink_b = AsyncMock(side_effect=RuntimeError("boom"))
        sink_c = AsyncMock()
        fanout.add_sink(sink_a)
        fanout.add_sink(sink_b)
        fanout.add_sink(sink_c)

        frame = _make_frame()
        await fanout.on_frame(frame)

        sink_a.assert_awaited_once_with(frame)
        sink_b.assert_awaited_once_with(frame)
        sink_c.assert_awaited_once_with(frame)

    @pytest.mark.asyncio
    async def test_multiple_frames(self) -> None:
        fanout = FrameFanout()
        sink = AsyncMock()
        fanout.add_sink(sink)

        frame1 = _make_frame(field="Soc", value=80)
        frame2 = _make_frame(field="BatteryLevel", value=78)
        await fanout.on_frame(frame1)
        await fanout.on_frame(frame2)

        assert sink.await_count == 2

    @pytest.mark.asyncio
    async def test_no_sinks_is_noop(self) -> None:
        fanout = FrameFanout()
        frame = _make_frame()
        await fanout.on_frame(frame)  # should not raise



"""Vendored protobuf message definitions for Tesla Vehicle Command Protocol.

These are minimal, hand-written message definitions that are wire-compatible
with Tesla's published `.proto` schemas.  Generated ``_pb2.py`` files can
replace them when full ``protoc`` generation is wired up.
"""



"""Async HTTP client for the Tesla Fleet API."""

from __future__ import annotations

import asyncio
import json
import os as _os
from typing import TYPE_CHECKING, Any

import httpx

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

from tescmd.api.errors import (
    AuthError,
    MissingScopesError,
    NetworkError,
    RateLimitError,
    RegistrationRequiredError,
    TeslaAPIError,
    VCPRequiredError,
    VehicleAsleepError,
)

# ---------------------------------------------------------------------------
# Region -> base URL mapping
# ---------------------------------------------------------------------------

# Allow a full override via TESLA_API_BASE_URL (e.g. for proxies / staging).
# When set the region parameter is ignored and all requests go to this URL.
_API_BASE_OVERRIDE: str | None = _os.environ.get("TESLA_API_BASE_URL")

REGION_BASE_URLS: dict[str, str] = {
    "na": "https://fleet-api.prd.na.vn.cloud.tesla.com",
    "eu": "https://fleet-api.prd.eu.vn.cloud.tesla.com",
    "cn": "https://fleet-api.prd.cn.vn.cloud.tesla.com",
}

# ---------------------------------------------------------------------------
# Tesla Fleet API documented rate limits (per device, per account).
# See: https://developer.tesla.com/docs/fleet-api/billing-and-limits
# ---------------------------------------------------------------------------

RATE_LIMITS = {
    "data": 60,  # Realtime data: 60 req/min
    "commands": 30,  # Device commands: 30 req/min
    "wakes": 3,  # Wakes: 3 req/min
    "auth": 20,  # Auth: 20 req/sec
}
_RATE_LIMIT_MAX_RETRIES = 3


class TeslaFleetClient:
    """Low-level async HTTP client for Tesla Fleet API endpoints."""

    def __init__(
        self,
        access_token: str,
        region: str = "na",
        timeout: float = 30.0,
        on_token_refresh: Callable[[], Awaitable[str | None]] | None = None,
        on_rate_limit_wait: Callable[[int, int, int], Awaitable[None]] | None = None,
    ) -> None:
        base_url = _API_BASE_OVERRIDE or REGION_BASE_URLS.get(region)
        if base_url is None:
            msg = f"Unknown region {region!r}; expected one of {sorted(REGION_BASE_URLS)}"
            raise ValueError(msg)

        self._access_token = access_token
        self._on_token_refresh = on_token_refresh
        self._on_rate_limit_wait = on_rate_limit_wait
        self._client = httpx.AsyncClient(
            base_url=base_url,
            timeout=timeout,
            headers={"Authorization": f"Bearer {access_token}"},
        )

    # -- public helpers -----------------------------------------------------

    async def get(self, path: str, **kwargs: Any) -> dict[str, Any]:
        """Issue a GET request and return the parsed JSON body."""
        return await self._request("GET", path, **kwargs)

    async def post(self, path: str, **kwargs: Any) -> dict[str, Any]:
        """Issue a POST request and return the parsed JSON body."""
        return await self._request("POST", path, **kwargs)

    async def delete(self, path: str, **kwargs: Any) -> dict[str, Any]:
        """Issue a DELETE request and return the parsed JSON body."""
        return await self._request("DELETE", path, **kwargs)

    def update_token(self, access_token: str) -> None:
        """Replace the current access token and update the header."""
        self._access_token = access_token
        self._client.headers["Authorization"] = f"Bearer {access_token}"

    async def close(self) -> None:
        """Close the underlying HTTP client."""
        await self._client.aclose()

    # -- async context manager -----------------------------------------------

    async def __aenter__(self) -> TeslaFleetClient:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: Any,
    ) -> None:
        await self.close()

    # -- internal ------------------------------------------------------------

    async def _request(
        self,
        method: str,
        path: str,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Issue an HTTP request with automatic retry on 429 rate-limit responses."""
        for attempt in range(_RATE_LIMIT_MAX_RETRIES + 1):
            try:
                return await self._send(method, path, **kwargs)
            except RateLimitError as exc:
                if attempt >= _RATE_LIMIT_MAX_RETRIES:
                    raise
                wait = exc.retry_after or min(2 ** (attempt + 1), 30)
                if self._on_rate_limit_wait:
                    await self._on_rate_limit_wait(wait, attempt + 1, _RATE_LIMIT_MAX_RETRIES)
                else:
                    await asyncio.sleep(wait)
        raise RateLimitError()  # unreachable, satisfies type checker

    async def _send(
        self,
        method: str,
        path: str,
        *,
        _retried: bool = False,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Send a single HTTP request, handling auth refresh on 401."""
        try:
            response = await self._client.request(method, path, **kwargs)
        except httpx.TimeoutException as exc:
            raise NetworkError(f"Request timed out: {exc}") from exc
        except httpx.ConnectError as exc:
            raise NetworkError(f"Connection error: {exc}") from exc

        # Handle 401 with optional token refresh
        if response.status_code == 401 and not _retried:
            if self._on_token_refresh is not None:
                new_token = await self._on_token_refresh()
                if new_token is not None:
                    self.update_token(new_token)
                    return await self._send(method, path, _retried=True, **kwargs)
            raise AuthError("Authentication failed", status_code=401)

        return self._parse_response(response)

    @staticmethod
    def _parse_response(response: httpx.Response) -> dict[str, Any]:
        """Translate HTTP status codes to domain exceptions and return JSON."""
        if response.status_code == 429:
            raw = response.headers.get("retry-after")
            retry_after: int | None = None
            if raw is not None:
                try:
                    retry_after = int(raw)
                except ValueError:
                    retry_after = 30  # Fallback for HTTP-date strings
            raise RateLimitError(retry_after=retry_after)

        if response.status_code == 408:
            # Tesla returns HTTP 408 for both "vehicle is asleep" and general
            # request timeouts.  Inspect the body to distinguish the two.
            body_text = response.text[:300]
            body_lower = body_text.lower()
            if "asleep" in body_lower or "offline" in body_lower or "unavailable" in body_lower:
                raise VehicleAsleepError("Vehicle is asleep", status_code=408)
            # No explicit sleep indicator — still raise VehicleAsleepError for
            # backwards compatibility (most 408s *are* sleep), but include the
            # raw body so callers can see what actually happened.
            msg = (
                f"Vehicle request timed out (HTTP 408): {body_text}"
                if body_text
                else "Vehicle is asleep"
            )
            raise VehicleAsleepError(msg, status_code=408)

        if response.status_code == 412:
            raise RegistrationRequiredError(
                "Your application is not registered with the Tesla Fleet API "
                "for this region. Run 'tescmd auth register' to fix this.",
                status_code=412,
            )

        if response.status_code == 403:
            text = response.text[:200]
            text_lower = text.lower()
            if "missing scopes" in text_lower:
                raise MissingScopesError(text, status_code=403)
            if "vehicle command protocol required" in text_lower:
                raise VCPRequiredError(
                    "This command is not available via the Fleet API when "
                    "Vehicle Command Protocol is enabled. Use the Tesla app "
                    "or run tescmd behind a tesla-http-proxy.",
                    status_code=403,
                )
            raise AuthError(f"HTTP 403: {text}", status_code=403)

        if response.status_code >= 400:
            text = response.text[:200]
            raise TeslaAPIError(
                f"HTTP {response.status_code}: {text}",
                status_code=response.status_code,
            )

        # 204 No Content — return empty dict (no body to parse)
        if response.status_code == 204:
            return {}

        try:
            result: dict[str, Any] = response.json()
        except json.JSONDecodeError as exc:
            raise TeslaAPIError(
                f"Invalid JSON response: {response.text[:200]}",
                status_code=response.status_code,
            ) from exc
        return result



"""CLI commands for EC key management (generate, deploy, validate, show, enroll)."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import click

from tescmd._internal.async_utils import run_async
from tescmd.cli._options import global_options
from tescmd.crypto.keys import (
    generate_ec_key_pair,
    get_key_fingerprint,
    get_public_key_path,
    has_key_pair,
    load_public_key_pem,
)
from tescmd.deploy.github_pages import (
    get_key_url,
    validate_key_url,
    wait_for_pages_deployment,
)
from tescmd.models.config import AppSettings

if TYPE_CHECKING:
    from tescmd.cli.main import AppContext
    from tescmd.output.formatter import OutputFormatter


# ---------------------------------------------------------------------------
# Command group
# ---------------------------------------------------------------------------

key_group = click.Group("key", help="EC key management")


# ---------------------------------------------------------------------------
# Commands
# ---------------------------------------------------------------------------


@key_group.command("generate")
@click.option("--force", is_flag=True, help="Overwrite existing keys")
@global_options
def generate_cmd(app_ctx: AppContext, force: bool) -> None:
    """Generate an EC P-256 key pair for Tesla Fleet API command signing."""
    run_async(_cmd_generate(app_ctx, force))


async def _cmd_generate(app_ctx: AppContext, force: bool) -> None:
    formatter = app_ctx.formatter
    settings = AppSettings()
    key_dir = Path(settings.config_dir).expanduser() / "keys"

    if has_key_pair(key_dir) and not force:
        if formatter.format == "json":
            formatter.output(
                {
                    "status": "exists",
                    "path": str(key_dir),
                    "fingerprint": get_key_fingerprint(key_dir),
                },
                command="key.generate",
            )
        else:
            formatter.rich.info(
                "[yellow]Key pair already exists.[/yellow] Use --force to overwrite."
            )
            formatter.rich.info(f"  Path: {key_dir}")
            formatter.rich.info(f"  Fingerprint: {get_key_fingerprint(key_dir)}")
        return

    priv_path, pub_path = generate_ec_key_pair(key_dir, overwrite=force)

    if formatter.format == "json":
        formatter.output(
            {
                "status": "generated",
                "private_key": str(priv_path),
                "public_key": str(pub_path),
                "fingerprint": get_key_fingerprint(key_dir),
            },
            command="key.generate",
        )
    else:
        formatter.rich.info("[green]Key pair generated.[/green]")
        formatter.rich.info(f"  Private key: {priv_path}")
        formatter.rich.info(f"  Public key:  {pub_path}")
        formatter.rich.info(f"  Fingerprint: {get_key_fingerprint(key_dir)}")


@key_group.command("deploy")
@click.option(
    "--repo",
    default=None,
    help="GitHub repo (e.g. user/user.github.io). Auto-detected if omitted.",
)
@click.option(
    "--method",
    type=click.Choice(["auto", "github", "tailscale"]),
    default="auto",
    help="Deployment method (default: auto-detect).",
)
@global_options
def deploy_cmd(app_ctx: AppContext, repo: str | None, method: str) -> None:
    """Deploy the public key to GitHub Pages or via Tailscale Funnel."""
    run_async(_cmd_deploy(app_ctx, repo, method=method))


async def _cmd_deploy(app_ctx: AppContext, repo: str | None, *, method: str = "auto") -> None:
    formatter = app_ctx.formatter
    settings = AppSettings()
    key_dir = Path(settings.config_dir).expanduser() / "keys"

    # Ensure keys exist
    if not has_key_pair(key_dir):
        if formatter.format == "json":
            formatter.output_error(
                code="no_keys",
                message="No key pair found. Run 'tescmd key generate' first.",
                command="key.deploy",
            )
        else:
            formatter.rich.error("No key pair found. Run [cyan]tescmd key generate[/cyan] first.")
        return

    # Resolve method
    resolved = _resolve_deploy_method(method, settings)

    if resolved == "tailscale":
        await _cmd_deploy_tailscale(formatter, key_dir)
    elif resolved == "github":
        await _cmd_deploy_github(formatter, settings, key_dir, repo)
    else:
        if formatter.format == "json":
            formatter.output_error(
                code="no_deploy_method",
                message=(
                    "No deployment method available. Install 'gh' CLI for GitHub Pages"
                    " or Tailscale for Funnel hosting."
                ),
                command="key.deploy",
            )
        else:
            formatter.rich.error(
                "No deployment method available. Install [cyan]gh[/cyan] CLI"
                " (GitHub Pages) or [cyan]tailscale[/cyan] (Funnel hosting)."
            )


def _resolve_deploy_method(method: str, settings: AppSettings) -> str | None:
    """Resolve 'auto' to a concrete method, or validate explicit choice."""
    if method in ("tailscale", "github"):
        return method

    # Auto-detect: GitHub Pages (always-on) > Tailscale (stable hostname)
    from tescmd.deploy.github_pages import is_gh_authenticated, is_gh_available

    if is_gh_available() and is_gh_authenticated():
        return "github"

    from tescmd.deploy.tailscale_serve import is_tailscale_serve_ready

    if run_async(is_tailscale_serve_ready()):
        return "tailscale"

    # Check saved hosting method as last resort
    if settings.hosting_method == "tailscale":
        return "tailscale"
    if settings.hosting_method == "github":
        return "github"

    return None


async def _cmd_deploy_tailscale(
    formatter: OutputFormatter,
    key_dir: Path,
) -> None:
    """Deploy public key via Tailscale Funnel."""
    from tescmd.deploy.tailscale_serve import (
        deploy_public_key_tailscale,
        get_key_url,
        is_tailscale_serve_ready,
        start_key_serving,
        wait_for_tailscale_deployment,
    )

    # Verify Tailscale is ready
    if not await is_tailscale_serve_ready():
        if formatter.format == "json":
            formatter.output_error(
                code="tailscale_not_ready",
                message=(
                    "Tailscale is not available or Funnel is not enabled."
                    " Ensure Tailscale is running and Funnel is enabled in your tailnet ACL."
                ),
                command="key.deploy",
            )
        else:
            formatter.rich.error(
                "Tailscale is not available or Funnel is not enabled.\n"
                "  Ensure Tailscale is running and Funnel is enabled in your tailnet ACL."
            )
        return

    if formatter.format != "json":
        formatter.rich.info("Deploying public key via Tailscale Funnel...")

    pem = load_public_key_pem(key_dir)
    await deploy_public_key_tailscale(pem)
    hostname = await start_key_serving()

    url = get_key_url(hostname)
    if formatter.format != "json":
        formatter.rich.info("[green]Tailscale serve + Funnel started.[/green]")
        formatter.rich.info(f"  URL: {url}")
        formatter.rich.info("")
        formatter.rich.info("Waiting for key to become accessible...")

    deployed = await wait_for_tailscale_deployment(hostname)

    if formatter.format == "json":
        formatter.output(
            {
                "status": "deployed" if deployed else "pending",
                "method": "tailscale",
                "hostname": hostname,
                "url": url,
                "accessible": deployed,
            },
            command="key.deploy",
        )
    elif deployed:
        formatter.rich.info("[green]Key is live and accessible.[/green]")
        formatter.rich.info("")
        formatter.rich.info(
            "[yellow]Note:[/yellow] The key is served by the Tailscale daemon."
            " If your machine is off or Tailscale stops,"
            " Tesla cannot reach your public key."
        )
    else:
        formatter.rich.info("[yellow]Key deployed but not yet accessible.[/yellow]")
        formatter.rich.info("  Run [cyan]tescmd key validate[/cyan] to check again later.")


async def _cmd_deploy_github(
    formatter: OutputFormatter,
    settings: AppSettings,
    key_dir: Path,
    repo: str | None,
) -> None:
    """Deploy public key via GitHub Pages."""
    from tescmd.deploy.github_pages import (
        create_pages_repo,
        deploy_public_key,
        get_gh_username,
        get_pages_domain,
        is_gh_authenticated,
        is_gh_available,
    )

    # Check gh CLI
    if not is_gh_available():
        if formatter.format == "json":
            formatter.output_error(
                code="gh_not_found",
                message="GitHub CLI (gh) is not installed. Install from https://cli.github.com",
                command="key.deploy",
            )
        else:
            formatter.rich.error(
                "GitHub CLI ([cyan]gh[/cyan]) is not installed."
                " Install from [link=https://cli.github.com]cli.github.com[/link]"
            )
        return

    if not is_gh_authenticated():
        if formatter.format == "json":
            formatter.output_error(
                code="gh_not_authenticated",
                message="GitHub CLI is not authenticated. Run 'gh auth login' first.",
                command="key.deploy",
            )
        else:
            formatter.rich.error(
                "GitHub CLI is not authenticated. Run [cyan]gh auth login[/cyan] first."
            )
        return

    # Determine repo
    repo_name: str | None = repo or settings.github_repo
    if not repo_name:
        username = get_gh_username()
        repo_name = f"{username}/{username}.github.io"
        if formatter.format != "json":
            formatter.rich.info(f"Using repo: [cyan]{repo_name}[/cyan]")

    # Create repo if needed and deploy
    if formatter.format != "json":
        formatter.rich.info("Creating repo if needed...")
    create_pages_repo(repo_name.split("/")[0])

    if formatter.format != "json":
        formatter.rich.info("Deploying public key...")

    pem = load_public_key_pem(key_dir)
    deploy_public_key(pem, repo_name)

    domain = get_pages_domain(repo_name)

    if formatter.format != "json":
        formatter.rich.info("[green]Key deployed.[/green]")
        formatter.rich.info(f"  URL: {get_key_url(domain)}")
        formatter.rich.info("")
        formatter.rich.info("Waiting for GitHub Pages to publish...")

    deployed = wait_for_pages_deployment(domain)

    if formatter.format == "json":
        formatter.output(
            {
                "status": "deployed" if deployed else "pending",
                "method": "github",
                "repo": repo_name,
                "domain": domain,
                "url": get_key_url(domain),
                "accessible": deployed,
            },
            command="key.deploy",
        )
    elif deployed:
        formatter.rich.info("[green]Key is live and accessible.[/green]")
    else:
        formatter.rich.info(
            "[yellow]Key deployed but not yet accessible."
            " GitHub Pages may still be building.[/yellow]"
        )
        formatter.rich.info("  Run [cyan]tescmd key validate[/cyan] to check again later.")


@key_group.command("validate")
@global_options
def validate_cmd(app_ctx: AppContext) -> None:
    """Check that the public key is accessible at the expected URL."""
    run_async(_cmd_validate(app_ctx))


async def _cmd_validate(app_ctx: AppContext) -> None:
    formatter = app_ctx.formatter
    settings = AppSettings()
    domain = settings.domain

    if not domain:
        if formatter.format == "json":
            formatter.output_error(
                code="no_domain",
                message=(
                    "TESLA_DOMAIN is not set. Set it in your .env file or run"
                    " 'tescmd setup' to configure."
                ),
                command="key.validate",
            )
        else:
            formatter.rich.error(
                "No domain configured. Run [cyan]tescmd setup[/cyan]"
                " or set TESLA_DOMAIN in your .env file."
            )
        return

    url = get_key_url(domain)
    accessible = validate_key_url(domain)

    if formatter.format == "json":
        formatter.output(
            {"url": url, "accessible": accessible, "domain": domain},
            command="key.validate",
        )
    elif accessible:
        formatter.rich.info(f"[green]Public key is accessible at:[/green] {url}")
    else:
        formatter.rich.info(f"[red]Public key NOT accessible at:[/red] {url}")
        formatter.rich.info("")
        formatter.rich.info("Possible causes:")
        formatter.rich.info("  - Key has not been deployed yet")
        formatter.rich.info("  - GitHub Pages is still building")
        formatter.rich.info("  - Domain is not configured correctly")
        formatter.rich.info("")
        formatter.rich.info("Run [cyan]tescmd key deploy[/cyan] to deploy your key.")


@key_group.command("show")
@global_options
def show_cmd(app_ctx: AppContext) -> None:
    """Display key path and fingerprint."""
    run_async(_cmd_show(app_ctx))


async def _cmd_show(app_ctx: AppContext) -> None:
    formatter = app_ctx.formatter
    settings = AppSettings()
    key_dir = Path(settings.config_dir).expanduser() / "keys"

    if not has_key_pair(key_dir):
        if formatter.format == "json":
            formatter.output(
                {"status": "not_found", "path": str(key_dir)},
                command="key.show",
            )
        else:
            formatter.rich.info(
                "No key pair found. Run [cyan]tescmd key generate[/cyan] to create one."
            )
        return

    pub_path = get_public_key_path(key_dir)
    fingerprint = get_key_fingerprint(key_dir)

    if formatter.format == "json":
        formatter.output(
            {
                "status": "found",
                "path": str(key_dir),
                "public_key": str(pub_path),
                "fingerprint": fingerprint,
            },
            command="key.show",
        )
    else:
        formatter.rich.info(f"Key directory: {key_dir}")
        formatter.rich.info(f"Public key:    {pub_path}")
        formatter.rich.info(f"Fingerprint:   {fingerprint}")

        domain = settings.domain
        if domain:
            formatter.rich.info(f"Expected URL:  {get_key_url(domain)}")


# ---------------------------------------------------------------------------
# Enroll command
# ---------------------------------------------------------------------------

# Tesla app portal URL for key enrollment.
# Initial key enrollment is NOT available via REST API or signed_command.
# Tesla's Go SDK explicitly blocks add_key_request for Fleet API connections
# (ErrRequiresBLE).  For Fleet API apps, enrollment happens through the
# Tesla app portal: the user opens https://tesla.com/_ak/<domain> on their
# phone, and the Tesla app handles the actual key pairing with the vehicle.
_TESLA_APP_KEY_URL = "https://tesla.com/_ak/{domain}"

# Tesla consent revocation URL.  Revoking consent removes the app's
# OAuth access and its virtual key from all vehicles on the account.
# There is no Fleet API endpoint to remove a virtual key — Tesla requires
# the owner to revoke access through their account or the vehicle itself.
_TESLA_CONSENT_REVOKE_URL = (
    "https://auth.tesla.com/user/revoke/consent"
    "?revoke_client_id={client_id}&back_url=https://tesla.com"
)


@key_group.command("enroll")
@click.option(
    "--open/--no-open",
    default=True,
    help="Open the enrollment URL in the default browser (default: open)",
)
@global_options
def enroll_cmd(
    app_ctx: AppContext,
    open: bool,
) -> None:
    """Enroll your EC key on a vehicle via the Tesla app.

    Opens the Tesla app enrollment page for your domain.  The vehicle
    owner approves the key in the Tesla app (Profile >
    Security & Privacy > Third-Party Apps).  Once approved, signed commands work
    automatically.
    """
    run_async(_cmd_enroll(app_ctx, open_browser=open))


async def _cmd_enroll(
    app_ctx: AppContext,
    *,
    open_browser: bool,
) -> None:
    import webbrowser

    formatter = app_ctx.formatter
    settings = AppSettings()
    key_dir = Path(settings.config_dir).expanduser() / "keys"

    # Step 1: Check key pair
    if not has_key_pair(key_dir):
        if formatter.format == "json":
            formatter.output_error(
                code="no_keys",
                message="No key pair found. Run 'tescmd key generate' first.",
                command="key.enroll",
            )
        else:
            formatter.rich.error("No key pair found. Run [cyan]tescmd key generate[/cyan] first.")
        raise SystemExit(1)

    fingerprint = get_key_fingerprint(key_dir)

    if formatter.format != "json":
        formatter.rich.info("[bold]Step 1: Checking key pair…[/bold]")
        formatter.rich.info(f"  [green]✓[/green] Key pair found (fingerprint: {fingerprint[:8]}…)")
        formatter.rich.info("")

    # Step 2: Check domain is configured
    domain = settings.domain
    if not domain:
        if formatter.format == "json":
            formatter.output_error(
                code="no_domain",
                message=(
                    "No domain configured. Run 'tescmd setup' or set TESLA_DOMAIN"
                    " in your .env file."
                ),
                command="key.enroll",
            )
        else:
            formatter.rich.error(
                "No domain configured. Run [cyan]tescmd setup[/cyan]"
                " or set TESLA_DOMAIN in your .env file."
            )
        raise SystemExit(1)

    if formatter.format != "json":
        formatter.rich.info("[bold]Step 2: Checking domain & public key…[/bold]")

    # Step 3: Verify public key is accessible
    key_accessible = validate_key_url(domain)
    key_url = get_key_url(domain)

    if not key_accessible:
        if formatter.format == "json":
            formatter.output_error(
                code="key_not_accessible",
                message=(
                    f"Public key not accessible at {key_url}. "
                    "Deploy with 'tescmd key deploy' first."
                ),
                command="key.enroll",
            )
        else:
            formatter.rich.error(f"Public key not accessible at [cyan]{key_url}[/cyan]")
            formatter.rich.info("  Deploy with [cyan]tescmd key deploy[/cyan] first.")
        raise SystemExit(1)

    if formatter.format != "json":
        formatter.rich.info(f"  [green]✓[/green] Domain: {domain}")
        formatter.rich.info(f"  [green]✓[/green] Public key accessible at {key_url}")
        formatter.rich.info("")

    # Step 4: Build enrollment URL and guide user
    enroll_url = _TESLA_APP_KEY_URL.format(domain=domain)

    if formatter.format == "json":
        formatter.output(
            {
                "status": "ready",
                "domain": domain,
                "fingerprint": fingerprint,
                "enroll_url": enroll_url,
                "key_url": key_url,
                "message": (
                    f"Open {enroll_url} on your phone."
                    " Tap 'Finish Setup', then approve 'Add Virtual Key' in the Tesla app."
                ),
            },
            command="key.enroll",
        )
        return

    formatter.rich.info("[bold]Step 3: Key enrollment[/bold]")
    formatter.rich.info("")
    formatter.rich.info("━" * 55)
    formatter.rich.info(
        "  [bold yellow]ACTION REQUIRED: Add virtual key in the Tesla app[/bold yellow]"
    )
    formatter.rich.info("")
    formatter.rich.info(f"  Enrollment URL: [link={enroll_url}]{enroll_url}[/link]")
    formatter.rich.info("")
    formatter.rich.info("  1. Scan the QR code on the page above with your phone[/bold]")
    formatter.rich.info("  2. The Tesla app will show an [bold]Add Virtual Key[/bold] prompt")
    formatter.rich.info("  3. Select all Scopes (if shown) and approve it")
    formatter.rich.info("")
    formatter.rich.info("  [dim]If the prompt doesn't appear, force-quit the Tesla app,[/dim]")
    formatter.rich.info("  [dim]and scan the QR code again.[/dim]")
    formatter.rich.info("━" * 55)
    formatter.rich.info("")

    if open_browser:
        formatter.rich.info("Opening enrollment URL in browser…")
        webbrowser.open(enroll_url)
        formatter.rich.info("")

    formatter.rich.info("After approving in the Tesla app, try a command:")
    formatter.rich.info("  [cyan]tescmd vehicle list[/cyan]")
    formatter.rich.info("  [cyan]tescmd charge status --wake[/cyan]")
    formatter.rich.info("")
    formatter.rich.info(
        "[dim]Tip: The QR code must be scanned on your phone"
        " that has the Tesla app installed.[/dim]"
    )


# ---------------------------------------------------------------------------
# Unenroll command
# ---------------------------------------------------------------------------


@key_group.command("unenroll")
@click.option(
    "--open/--no-open",
    default=True,
    help="Open the revocation URL in the default browser (default: open)",
)
@global_options
def unenroll_cmd(
    app_ctx: AppContext,
    open: bool,
) -> None:
    """Remove your virtual key and revoke app access.

    Shows instructions for removing the tescmd virtual key from your
    vehicle(s) and optionally opens the Tesla consent revocation page
    to revoke OAuth access entirely.
    """
    run_async(_cmd_unenroll(app_ctx, open_browser=open))


async def _cmd_unenroll(
    app_ctx: AppContext,
    *,
    open_browser: bool,
) -> None:
    import webbrowser

    formatter = app_ctx.formatter
    settings = AppSettings()
    client_id = settings.client_id

    # Build revocation URL if client_id is available
    revoke_url: str | None = None
    if client_id:
        revoke_url = _TESLA_CONSENT_REVOKE_URL.format(client_id=client_id)

    if formatter.format == "json":
        data: dict[str, object] = {
            "status": "instructions",
            "revoke_url": revoke_url,
            "methods": [
                {
                    "name": "vehicle_touchscreen",
                    "steps": "Controls > Locks > tap trash icon on key > scan key card",
                    "speed": "immediate",
                },
                {
                    "name": "tesla_app",
                    "steps": "Profile > Security & Privacy > Third-Party Apps > Remove",
                    "speed": "up to 2 hours",
                },
                {
                    "name": "tesla_account_web",
                    "steps": "accounts.tesla.com > Security > Third Party Apps > Manage",
                    "speed": "up to 2 hours",
                },
            ],
            "message": (
                "Remove the virtual key via the vehicle touchscreen, Tesla app, "
                "or accounts.tesla.com. "
                + (
                    f"Revoke OAuth access at {revoke_url}"
                    if revoke_url
                    else "Set TESLA_CLIENT_ID to generate a consent revocation URL."
                )
            ),
        }
        formatter.output(data, command="key.unenroll")
        return

    formatter.rich.info("[bold]How to remove the tescmd virtual key[/bold]")
    formatter.rich.info("")
    formatter.rich.info("━" * 55)
    formatter.rich.info("")

    # Method 1: Vehicle touchscreen (immediate)
    formatter.rich.info("  [bold]Option 1: Vehicle touchscreen[/bold] [green](immediate)[/green]")
    formatter.rich.info("  1. On the vehicle touchscreen, tap [bold]Controls > Locks[/bold]")
    formatter.rich.info("  2. Find the tescmd key in the key list")
    formatter.rich.info("  3. Tap the [bold]trash icon[/bold] next to it")
    formatter.rich.info("  4. Scan your [bold]key card[/bold] on the card reader to confirm")
    formatter.rich.info("")

    # Method 2: Tesla app
    formatter.rich.info(
        "  [bold]Option 2: Tesla app[/bold] [yellow](may take up to 2 hours)[/yellow]"
    )
    formatter.rich.info("  1. Open the Tesla app > [bold]Profile > Security & Privacy[/bold]")
    formatter.rich.info("  2. Tap [bold]Third-Party Apps[/bold]")
    formatter.rich.info("  3. Find tescmd and tap [bold]Remove[/bold]")
    formatter.rich.info("")

    # Method 3: Tesla account website
    formatter.rich.info(
        "  [bold]Option 3: Tesla account website[/bold] [yellow](may take up to 2 hours)[/yellow]"
    )
    formatter.rich.info(
        "  1. Sign in at [link=https://accounts.tesla.com]accounts.tesla.com[/link]"
    )
    formatter.rich.info("  2. Go to [bold]Security > Third Party Apps[/bold]")
    formatter.rich.info("  3. Find tescmd and tap [bold]Manage > Remove[/bold]")
    formatter.rich.info("")
    formatter.rich.info("━" * 55)
    formatter.rich.info("")

    # OAuth consent revocation
    if revoke_url:
        formatter.rich.info("[bold]Revoke OAuth access entirely[/bold]")
        formatter.rich.info("")
        formatter.rich.info("  To also revoke this app's OAuth access to your Tesla account:")
        formatter.rich.info(f"  [link={revoke_url}]{revoke_url}[/link]")
        formatter.rich.info("")

        if open_browser:
            formatter.rich.info("Opening revocation page in browser…")
            webbrowser.open(revoke_url)
            formatter.rich.info("")
    else:
        formatter.rich.info("[dim]Set TESLA_CLIENT_ID to generate a consent revocation URL.[/dim]")
        formatter.rich.info("")

    # Cleanup guidance
    formatter.rich.info("[bold]Local cleanup[/bold]")
    formatter.rich.info("")
    formatter.rich.info("  To also remove local credentials and keys:")
    formatter.rich.info("  [cyan]tescmd auth logout[/cyan]     — clear stored OAuth tokens")
    formatter.rich.info("  [cyan]tescmd cache clear[/cyan]     — clear cached API responses")
    formatter.rich.info(
        f"  [dim]Key files are stored in: {Path(settings.config_dir).expanduser() / 'keys'}[/dim]"
    )
